from __future__ import annotations

import json
from asyncio import gather, get_running_loop
from csv import DictReader
from datetime import datetime
from functools import partial
from pathlib import Path
from sys import getsizeof
from typing import Any, AsyncGenerator
from urllib.parse import urlencode

import aiofiles
from httpx import AsyncClient

from meilisearch_python_async._http_requests import HttpRequests
from meilisearch_python_async.errors import InvalidDocumentError, MeiliSearchError, PayloadTooLarge
from meilisearch_python_async.models.index import IndexStats
from meilisearch_python_async.models.search import SearchResults
from meilisearch_python_async.models.settings import MeiliSearchSettings
from meilisearch_python_async.models.task import TaskId, TaskStatus
from meilisearch_python_async.task import wait_for_task


class Index:
    """Index class gives access to all indexes routes and child routes.

    https://docs.meilisearch.com/reference/api/indexes.html
    """

    def __init__(
        self,
        http_client: AsyncClient,
        uid: str,
        primary_key: str | None = None,
        created_at: str | datetime | None = None,
        updated_at: str | datetime | None = None,
    ):
        """Class initializer.

        **Args:**

        * **http_client:** An instance of the AsyncClient. This automatically gets passed by the
            Client when creating and Index instance.
        * **uid:** The index's unique identifier.
        * **primary_key:** The primary key of the documents. Defaults to None.
        * **created_at:** The date and time the index was created. Defaults to None.
        * **updated_at:** The date and time the index was last updated. Defaults to None.
        """
        self.uid = uid
        self.primary_key = primary_key
        self.created_at: datetime | None = Index._iso_to_date_time(created_at)
        self.updated_at: datetime | None = Index._iso_to_date_time(updated_at)
        self._base_url = "indexes/"
        self._base_url_with_uid = f"{self._base_url}{self.uid}"
        self._documents_url = f"{self._base_url_with_uid}/documents"
        self._stats_url = f"{self._base_url_with_uid}/stats"
        self._settings_url = f"{self._base_url_with_uid}/settings"
        self.http_client = http_client
        self._http_requests = HttpRequests(http_client)

    def __str__(self) -> str:
        return f"{type(self).__name__}(uid={self.uid}, primary_key={self.primary_key}, created_at={self.created_at}, updated_at={self.updated_at})"

    def __repr__(self) -> str:
        return f"{type(self).__name__}(uid={self.uid!r}, primary_key={self.primary_key!r}, created_at={self.created_at!r}, updated_at={self.updated_at!r})"

    async def delete(self) -> TaskStatus:
        """Deletes the index.

        **Returns:** The details of the task status.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        **

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.delete()
        ```
        """
        url = f"{self._base_url_with_uid}"
        response = await self._http_requests.delete(url)
        return TaskStatus(**response.json())

    async def delete_if_exists(self) -> bool:
        """Delete the index if it already exists.

        **Returns:** True if the index was deleted or False if not.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.delete_if_exists()
        ```
        """
        response = await self.delete()
        status = await wait_for_task(self.http_client, response.uid)
        if status.status == "succeeded":
            return True

        return False

    async def update(self, primary_key: str) -> Index:
        """Update the index primary key.

        **Args:**

        * **primary_key:** The primary key of the documents.

        **Returns:** An instance of the Index with the updated information.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     updated_index = await index.update()
        ```
        """
        payload = {"primaryKey": primary_key}
        url = f"{self._base_url_with_uid}"
        response = await self._http_requests.put(url, payload)
        await wait_for_task(self.http_client, response.json()["uid"], timeout_in_ms=100000)
        index_response = await self._http_requests.get(f"{url}")
        self.primary_key = index_response.json()["primaryKey"]
        return self

    async def fetch_info(self) -> Index:
        """Gets the infromation about the index.

        **Returns:** An instance of the Index containing the retrieved information.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     index_info = await index.fetch_info()
        ```
        """
        url = f"{self._base_url_with_uid}"
        response = await self._http_requests.get(url)
        index_dict = response.json()
        self.primary_key = index_dict["primaryKey"]
        loop = get_running_loop()
        self.created_at = await loop.run_in_executor(
            None, partial(Index._iso_to_date_time, index_dict["createdAt"])
        )
        self.updated_at = await loop.run_in_executor(
            None, partial(Index._iso_to_date_time, index_dict["updatedAt"])
        )
        return self

    async def get_primary_key(self) -> str | None:
        """Get the primary key.

        **Returns:** The primary key for the documents in the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     primary_key = await index.get_primary_key()
        ```
        """
        info = await self.fetch_info()
        return info.primary_key

    @classmethod
    async def create(
        cls, http_client: AsyncClient, uid: str, primary_key: str | None = None
    ) -> Index:
        """Creates a new index.

        In general this method should not be used directly and instead the index should be created
        through the `Client`.

        **Args:**

        * **http_client:** An instance of the AsyncClient. This automatically gets passed by the
            Client when creating and Index instance.
        * **uid:** The index's unique identifier.
        * **primary_key:** The primary key of the documents. Defaults to None.

        **Returns:** An instance of Index containing the information of the newly created index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = await index.create(client, "movies")
        ```
        """
        if not primary_key:
            payload = {"uid": uid}
        else:
            payload = {"primaryKey": primary_key, "uid": uid}

        url = "indexes"
        http_request = HttpRequests(http_client)
        response = await http_request.post(url, payload)
        await wait_for_task(http_client, response.json()["uid"], timeout_in_ms=100000)
        index_response = await http_request.get(f"{url}/{uid}")
        index_dict = index_response.json()
        return cls(
            http_client=http_client,
            uid=index_dict["uid"],
            primary_key=index_dict["primaryKey"],
            created_at=index_dict["createdAt"],
            updated_at=index_dict["updatedAt"],
        )

    async def get_stats(self) -> IndexStats:
        """Get stats of the index.

        **Returns:** Stats of the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     stats = await index.get_stats()
        ```
        """
        url = f"{self._stats_url}"
        response = await self._http_requests.get(url)

        return IndexStats(**response.json())

    async def search(
        self,
        query: str | None = None,
        *,
        offset: int = 0,
        limit: int = 20,
        filter: str | list[str | list[str]] | None = None,
        facets_distribution: list[str] | None = None,
        attributes_to_retrieve: list[str] = ["*"],
        attributes_to_crop: list[str] | None = None,
        crop_length: int = 200,
        attributes_to_highlight: list[str] | None = None,
        sort: list[str] | None = None,
        matches: bool = False,
    ) -> SearchResults:
        """Search the index.

        **Args:**

        * **query:** String containing the word(s) to search
        * **offset:** Number of documents to skip. Defaults to 0.
        * **limit:** Maximum number of documents returned. Defaults to 20.
        * **filter:** Filter queries by an attribute value. Defaults to None.
        * **facets_distribution:** Facets for which to retrieve the matching count. Defaults to None.
        * **attributes_to_retrieve:** Attributes to display in the returned documents.
            Defaults to ["*"].
        * **attributes_to_crop:** Attributes whose values have to be cropped. Defaults to None.
        * **crop_length:** Length used to crop field values. Defaults to 200.
        * **attributes_to_highlight:** Attributes whose values will contain highlighted matching terms.
            Defaults to None.
        * **sort:** Attributes by which to sort the results. Defaults to None.
        * **matches:** Defines whether an object that contains information about the matches should be
            returned or not. Defaults to False.

        **Returns:** Results of the search

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     search_results = await index.search("Tron")
        ```
        """
        body = {
            "q": query,
            "offset": offset,
            "limit": limit,
            "filter": filter,
            "facetsDistribution": facets_distribution,
            "attributesToRetrieve": attributes_to_retrieve,
            "attributesToCrop": attributes_to_crop,
            "cropLength": crop_length,
            "attributesToHighlight": attributes_to_highlight,
            "sort": sort,
            "matches": matches,
        }
        url = f"{self._base_url_with_uid}/search"
        response = await self._http_requests.post(url, body=body)

        return SearchResults(**response.json())

    async def get_document(self, document_id: str) -> dict:
        """Get one document with given document identifier.

        **Args:**

        * **document_id:** Unique identifier of the document.

        **Returns:** The document information

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     document = await index.get_document("1234")
        ```
        """
        url = f"{self._documents_url}/{document_id}"
        response = await self._http_requests.get(url)
        return response.json()

    async def get_documents(
        self, *, offset: int = 0, limit: int = 20, attributes_to_retrieve: list[str] | None = None
    ) -> list[dict[str, Any]] | None:
        """Get a batch documents from the index.

        **Args:**

        * **offset:** Number of documents to skip. Defaults to 0.
        * **limit:** Maximum number of documents returnedd. Defaults to 20.
        * **attributes_to_retrieve:** Document attributes to show. If this value is None then all
            attributes are retrieved. Defaults to None.

        **Returns:** A list of documents.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.


        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     documents = await index.get_documents()
        ```
        """
        parameters: dict[str, Any] = {
            "offset": offset,
            "limit": limit,
        }

        if attributes_to_retrieve:
            parameters["attributesToRetrieve"] = ",".join(attributes_to_retrieve)

        url = f"{self._documents_url}?{urlencode(parameters)}"
        response = await self._http_requests.get(url)

        if not response.json():
            return None

        return response.json()

    async def add_documents(self, documents: list[dict], primary_key: str | None = None) -> TaskId:
        """Add documents to the index.

        **Args:**

        * **documents:** List of documents.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> >>> documents = [
        >>>     {"id": 1, "title": "Movie 1", "genre": "comedy"},
        >>>     {"id": 2, "title": "Movie 2", "genre": "drama"},
        >>> ]
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents(documents)
        ```
        """
        url = f"{self._documents_url}"
        if primary_key:
            formatted_primary_key = urlencode({"primaryKey": primary_key})
            url = f"{url}?{formatted_primary_key}"

        response = await self._http_requests.post(url, documents)
        return TaskId(**response.json())

    async def add_documents_auto_batch(
        self,
        documents: list[dict],
        *,
        max_payload_size: int = 104857600,
        primary_key: str | None = None,
    ) -> list[TaskId]:
        """Automatically splits the documents into batches when adding.

        Each batch tries to fill the max_payload_size

        **Args:**

        * **documents:** List of documents.
        * **max_payload_size:** The maximum payload size in bytes. Defaults to 104857600 (100MB).
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.
            PayloadTooLarge: If the largest document is larget than the max_payload_size

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> >>> documents = [
        >>>     {"id": 1, "title": "Movie 1", "genre": "comedy"},
        >>>     {"id": 2, "title": "Movie 2", "genre": "drama"},
        >>> ]
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_auto_batch(documents)
        ```
        """
        task_ids = []
        async for batch in Index._generate_auto_batches(documents, max_payload_size):
            task_id = await self.add_documents(batch, primary_key)
            task_ids.append(task_id)

        return task_ids

    async def add_documents_in_batches(
        self, documents: list[dict], *, batch_size: int = 1000, primary_key: str | None = None
    ) -> list[TaskId]:
        """Adds documents in batches to reduce RAM usage with indexing.

        **Args:**

        * **documents:** List of documents.
        * **batch_size:** The number of documents that should be included in each batch.
            Defaults to 1000.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> >>> documents = [
        >>>     {"id": 1, "title": "Movie 1", "genre": "comedy"},
        >>>     {"id": 2, "title": "Movie 2", "genre": "drama"},
        >>> ]
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_in_batches(documents)
        ```
        """
        task_ids: list[TaskId] = []

        async for document_batch in Index._batch(documents, batch_size):
            task_id = await self.add_documents(document_batch, primary_key)
            task_ids.append(task_id)

        return task_ids

    async def add_documents_from_directory(
        self,
        directory_path: Path | str,
        *,
        primary_key: str | None = None,
        document_type: str = "json",
        combine_documents: bool = True,
    ) -> list[TaskId]:
        """Load all json files from a directory and add the documents to the index.

        **Args:**

        * **directory_path:** Path to the directory that contains the json files.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.
        * **document_type:** The type of document being added. Accepted types are json, csv, and
                ndjson. For csv files the first row of the document should be a header row contining
                the field names, and ever for should have a title.
        * **combine_documents:** If set to True this will combine the documents from all the files
            before indexing them. Defaults to True.

        **Returns:** Update id to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> directory_path = Path("/path/to/directory/containing/files")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_from_directory(directory_path)
        ```
        """
        directory = Path(directory_path) if isinstance(directory_path, str) else directory_path

        if combine_documents:
            all_documents = []
            for path in directory.iterdir():
                if path.suffix == f".{document_type}":
                    documents = await Index._load_documents_from_file(path)
                    all_documents.append(documents)

            Index._raise_on_no_documents(all_documents, document_type, directory_path)

            loop = get_running_loop()
            combined = await loop.run_in_executor(
                None, partial(Index._combine_documents, all_documents)
            )

            response = await self.add_documents(combined, primary_key)
            return [response]

        add_documents = []
        for path in directory.iterdir():
            if path.suffix == f".{document_type}":
                documents = await Index._load_documents_from_file(path)
                add_documents.append(self.add_documents(documents, primary_key))

        Index._raise_on_no_documents(add_documents, document_type, directory_path)

        if len(add_documents) > 1:
            # Send the first document on its own before starting the gather. Otherwise MeiliSearch
            # returns an error because it thinks all entries are trying to create the same index.
            first_response = [await add_documents.pop()]
            responses = await gather(*add_documents)
            responses = [*first_response, *responses]
        else:
            responses = [await add_documents[0]]

        return responses

    async def add_documents_from_directory_auto_batch(
        self,
        directory_path: Path | str,
        *,
        max_payload_size: int = 104857600,
        primary_key: str | None = None,
        document_type: str = "json",
        combine_documents: bool = True,
    ) -> list[TaskId]:
        """Load all json files from a directory and add the documents to the index.

        Documents are automatically split into batches as large as possible based on the max payload
        size.

        **Args:**

        * **directory_path:** Path to the directory that contains the json files.
        * **max_payload_size:** The maximum payload size in bytes. Defaults to 104857600.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.
        * **document_type:** The type of document being added. Accepted types are json, csv, and
                ndjson. For csv files the first row of the document should be a header row contining
                the field names, and ever for should have a title.
        * **combine_documents:** If set to True this will combine the documents from all the files
            before indexing them. Defaults to True.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> directory_path = Path("/path/to/directory/containing/files")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_from_directory_auto_batch(directory_path)
        ```
        """
        directory = Path(directory_path) if isinstance(directory_path, str) else directory_path

        if combine_documents:
            all_documents = []
            for path in directory.iterdir():
                if path.suffix == f".{document_type}":
                    documents = await Index._load_documents_from_file(path)
                    all_documents.append(documents)

            Index._raise_on_no_documents(all_documents, document_type, directory_path)

            loop = get_running_loop()
            combined = await loop.run_in_executor(
                None, partial(Index._combine_documents, all_documents)
            )

            return await self.add_documents_auto_batch(
                combined, max_payload_size=max_payload_size, primary_key=primary_key
            )

        responses: list[TaskId] = []

        add_documents = []
        for path in directory.iterdir():
            if path.suffix == f".{document_type}":
                documents = await Index._load_documents_from_file(path)
                add_documents.append(
                    self.add_documents_auto_batch(
                        documents, max_payload_size=max_payload_size, primary_key=primary_key
                    )
                )

        Index._raise_on_no_documents(add_documents, document_type, directory_path)

        if len(add_documents) > 1:
            # Send the first document on its own before starting the gather. Otherwise MeiliSearch
            # returns an error because it thinks all entries are trying to create the same index.
            first_response = await add_documents.pop()
            responses_gather = await gather(*add_documents)
            responses = [*first_response, *[x for y in responses_gather for x in y]]
        else:
            responses = await add_documents[0]

        return responses

    async def add_documents_from_directory_in_batches(
        self,
        directory_path: Path | str,
        *,
        batch_size: int = 1000,
        primary_key: str | None = None,
        document_type: str = "json",
        combine_documents: bool = True,
    ) -> list[TaskId]:
        """Load all json files from a directory and add the documents to the index in batches.

        **Args:**

        * **directory_path:** Path to the directory that contains the json files.
        * **batch_size:** The number of documents that should be included in each batch.
            Defaults to 1000.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.
        * **document_type:** The type of document being added. Accepted types are json, csv, and
            ndjson. For csv files the first row of the document should be a header row contining
            the field names, and ever for should have a title.
        * **combine_documents:** If set to True this will combine the documents from all the files
            before indexing them. Defaults to True.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> directory_path = Path("/path/to/directory/containing/files")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_from_directory_in_batches(directory_path)
        ```
        """
        directory = Path(directory_path) if isinstance(directory_path, str) else directory_path

        if combine_documents:
            all_documents = []
            for path in directory.iterdir():
                if path.suffix == f".{document_type}":
                    documents = await Index._load_documents_from_file(path)
                    all_documents.append(documents)

            Index._raise_on_no_documents(all_documents, document_type, directory_path)

            loop = get_running_loop()
            combined = await loop.run_in_executor(
                None, partial(Index._combine_documents, all_documents)
            )

            return await self.add_documents_in_batches(
                combined, batch_size=batch_size, primary_key=primary_key
            )

        responses: list[TaskId] = []

        add_documents = []
        for path in directory.iterdir():
            if path.suffix == f".{document_type}":
                documents = await Index._load_documents_from_file(path)
                add_documents.append(
                    self.add_documents_in_batches(
                        documents, batch_size=batch_size, primary_key=primary_key
                    )
                )

        Index._raise_on_no_documents(add_documents, document_type, directory_path)

        if len(add_documents) > 1:
            # Send the first document on its own before starting the gather. Otherwise MeiliSearch
            # returns an error because it thinks all entries are trying to create the same index.
            first_response = await add_documents.pop()
            responses_gather = await gather(*add_documents)
            responses = [*first_response, *[x for y in responses_gather for x in y]]
        else:
            responses = await add_documents[0]

        return responses

    async def add_documents_from_file(
        self, file_path: Path | str, primary_key: str | None = None
    ) -> TaskId:
        """Add documents to the index from a json file.

        **Args:**

        * **file_path:** Path to the json file.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** Update id to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> file_path = Path("/path/to/file.json")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_from_file(file_path)
        ```
        """
        documents = await Index._load_documents_from_file(file_path)

        return await self.add_documents(documents, primary_key=primary_key)

    async def add_documents_from_file_auto_batch(
        self,
        file_path: Path | str,
        *,
        max_payload_size: int = 104857600,
        primary_key: str | None = None,
    ) -> list[TaskId]:
        """Automatically splits the documents into batches when adding from a json file.

        Each batch tries to fill the max_payload_size

        **Args:**

        * **file_path:** Path to the json file.
        * **max_payload_size:** The maximum payload size in bytes. Defaults to 104857600 (100MB).
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid.
        * **MeiliSearchCommunicationError:** If there was an error communicating with the server.
        * **MeiliSearchApiError:** If the MeiliSearch API returned an error.
        * **PayloadTooLarge:** If the largest document is larget than the max_payload_size

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> file_path = Path("/path/to/file.json")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_from_file_auto_batch(file_path)
        ```
        """
        documents = await Index._load_documents_from_file(file_path)

        task_ids = []
        async for batch in Index._generate_auto_batches(documents, max_payload_size):
            task_id = await self.add_documents(batch, primary_key)
            task_ids.append(task_id)

        return task_ids

    async def add_documents_from_file_in_batches(
        self, file_path: Path | str, *, batch_size: int = 1000, primary_key: str | None = None
    ) -> list[TaskId]:
        """Adds documents form a json file in batches to reduce RAM usage with indexing.

        **Args:**

        * **file_path:** Path to the json file.
        * **batch_size:** The number of documents that should be included in each batch.
            Defaults to 1000.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> file_path = Path("/path/to/file.json")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_from_file_in_batches(file_path)
        ```
        """
        documents = await Index._load_documents_from_file(file_path)

        return await self.add_documents_in_batches(
            documents, batch_size=batch_size, primary_key=primary_key
        )

    async def add_documents_from_raw_file(
        self, file_path: Path | str, primary_key: str | None = None
    ) -> TaskId:
        """Directly send csv or ndjson files to MeiliSearch without pre-processing.

        The can reduce RAM usage from MeiliSearch during indexing, but does not include the option
        for batching.

        **Args:**

        * **file_path:** The path to the file to send to MeiliSearch. Only csv and ndjson files are
            allowed.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** Update id to track the action.

        **Raises:**

        * **ValueError:** If the file is not a csv or ndjson file.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> file_path = Path("/path/to/file.csv")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.add_documents_from_raw_file(file_path)
        ```
        """
        upload_path = Path(file_path) if isinstance(file_path, str) else file_path
        if not upload_path.exists():
            raise MeiliSearchError("No file found at the specified path")

        if upload_path.suffix not in (".csv", ".ndjson"):
            raise ValueError("Only csv and ndjson files can be sent as binary files")

        content_type = "text/csv" if upload_path.suffix == ".csv" else "application/x-ndjson"
        url = self._documents_url
        if primary_key:
            formatted_primary_key = urlencode({"primaryKey": primary_key})
            url = f"{url}?{formatted_primary_key}"

        async with aiofiles.open(upload_path, "r") as f:
            data = await f.read()

        response = await self._http_requests.post(url, body=data, content_type=content_type)

        return TaskId(**response.json())

    async def update_documents(
        self, documents: list[dict], primary_key: str | None = None
    ) -> TaskId:
        """Update documents in the index.

        **Args:**

        * **documents:** List of documents.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> documents = [
        >>>     {"id": 1, "title": "Movie 1", "genre": "comedy"},
        >>>     {"id": 2, "title": "Movie 2", "genre": "drama"},
        >>> ]
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents(documents)
        ```
        """
        url = f"{self._documents_url}"
        if primary_key:
            formatted_primary_key = urlencode({"primaryKey": primary_key})
            url = f"{url}?{formatted_primary_key}"

        response = await self._http_requests.put(url, documents)
        return TaskId(**response.json())

    async def update_documents_auto_batch(
        self,
        documents: list[dict],
        *,
        max_payload_size: int = 104857600,
        primary_key: str | None = None,
    ) -> list[TaskId]:
        """Automatically splits the documents into batches when updating.

        Each batch tries to fill the max_payload_size

        **Args:**

        * **documents:** List of documents.
        * **max_payload_size:** The maximum payload size in bytes. Defaults to 104857600 (100MB).
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.
            PayloadTooLarge: If the largest document is larget than the max_payload_size

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> documents = [
        >>>     {"id": 1, "title": "Movie 1", "genre": "comedy"},
        >>>     {"id": 2, "title": "Movie 2", "genre": "drama"},
        >>> ]
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_auto_batch(documents)
        ```
        """
        task_ids = []
        async for batch in Index._generate_auto_batches(documents, max_payload_size):
            task_id = await self.update_documents(batch, primary_key)
            task_ids.append(task_id)

        return task_ids

    async def update_documents_in_batches(
        self, documents: list[dict], *, batch_size: int = 1000, primary_key: str | None = None
    ) -> list[TaskId]:
        """Update documents in batches to reduce RAM usage with indexing.

        Each batch tries to fill the max_payload_size

        **Args:**

        * **documents:** List of documents.
        * **batch_size:** The number of documents that should be included in each batch.
            Defaults to 1000.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> documents = [
        >>>     {"id": 1, "title": "Movie 1", "genre": "comedy"},
        >>>     {"id": 2, "title": "Movie 2", "genre": "drama"},
        >>> ]
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_in_batches(documents)
        ```
        """
        task_ids: list[TaskId] = []

        async for document_batch in Index._batch(documents, batch_size):
            task_id = await self.update_documents(document_batch, primary_key)
            task_ids.append(task_id)

        return task_ids

    async def update_documents_from_directory(
        self,
        directory_path: Path | str,
        *,
        primary_key: str | None = None,
        document_type: str = "json",
        combine_documents: bool = True,
    ) -> list[TaskId]:
        """Load all json files from a directory and update the documents.

        **Args:**

        * **directory_path:** Path to the directory that contains the json files.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.
        * **document_type:** The type of document being added. Accepted types are json, csv, and
                ndjson. For csv files the first row of the document should be a header row contining
                the field names, and ever for should have a title.
        * **combine_documents:** If set to True this will combine the documents from all the files
            before indexing them. Defaults to True.

        **Returns:** Update id to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> directory_path = Path("/path/to/directory/containing/files")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_from_directory(directory_path)
        ```
        """
        directory = Path(directory_path) if isinstance(directory_path, str) else directory_path

        if combine_documents:
            all_documents = []
            for path in directory.iterdir():
                if path.suffix == f".{document_type}":
                    documents = await Index._load_documents_from_file(path)
                    all_documents.append(documents)

            Index._raise_on_no_documents(all_documents, document_type, directory_path)

            loop = get_running_loop()
            combined = await loop.run_in_executor(
                None, partial(Index._combine_documents, all_documents)
            )

            response = await self.update_documents(combined, primary_key)
            return [response]

        update_documents = []
        for path in directory.iterdir():
            if path.suffix == f".{document_type}":
                documents = await Index._load_documents_from_file(path)
                update_documents.append(self.update_documents(documents, primary_key))

        Index._raise_on_no_documents(update_documents, document_type, directory_path)

        if len(update_documents) > 1:
            # Send the first document on its own before starting the gather. Otherwise MeiliSearch
            # returns an error because it thinks all entries are trying to create the same index.
            first_response = [await update_documents.pop()]
            responses = await gather(*update_documents)
            responses = [*first_response, *responses]
        else:
            responses = [await update_documents[0]]

        return responses

    async def update_documents_from_directory_auto_batch(
        self,
        directory_path: Path | str,
        *,
        max_payload_size: int = 104857600,
        primary_key: str | None = None,
        document_type: str = "json",
        combine_documents: bool = True,
    ) -> list[TaskId]:
        """Load all json files from a directory and update the documents.

        Documents are automatically split into batches as large as possible based on the max payload
        size.

        **Args:**

        * **directory_path:** Path to the directory that contains the json files.
        * **max_payload_size:** The maximum payload size in bytes. Defaults to 104857600.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.
        * **document_type:** The type of document being added. Accepted types are json, csv, and
                ndjson. For csv files the first row of the document should be a header row contining
                the field names, and ever for should have a title.
        * **combine_documents:** If set to True this will combine the documents from all the files
            before indexing them. Defaults to True.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> directory_path = Path("/path/to/directory/containing/files")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_from_directory_auto_batch(directory_path)
        ```
        """
        directory = Path(directory_path) if isinstance(directory_path, str) else directory_path

        if combine_documents:
            all_documents = []
            for path in directory.iterdir():
                if path.suffix == f".{document_type}":
                    documents = await Index._load_documents_from_file(path)
                    all_documents.append(documents)

            Index._raise_on_no_documents(all_documents, document_type, directory_path)

            loop = get_running_loop()
            combined = await loop.run_in_executor(
                None, partial(Index._combine_documents, all_documents)
            )

            return await self.update_documents_auto_batch(
                combined, max_payload_size=max_payload_size, primary_key=primary_key
            )

        update_documents = []
        for path in directory.iterdir():
            if path.suffix == f".{document_type}":
                documents = await Index._load_documents_from_file(path)
                update_documents.append(
                    self.update_documents_auto_batch(
                        documents, max_payload_size=max_payload_size, primary_key=primary_key
                    )
                )

        Index._raise_on_no_documents(update_documents, document_type, directory_path)

        if len(update_documents) > 1:
            # Send the first document on its own before starting the gather. Otherwise MeiliSearch
            # returns an error because it thinks all entries are trying to create the same index.
            first_response = await update_documents.pop()
            responses = await gather(*update_documents)
            responses = [*first_response, *[x for y in responses for x in y]]
        else:
            responses = await update_documents[0]

        return responses

    async def update_documents_from_directory_in_batches(
        self,
        directory_path: Path | str,
        *,
        batch_size: int = 1000,
        primary_key: str | None = None,
        document_type: str = "json",
        combine_documents: bool = True,
    ) -> list[TaskId]:
        """Load all json files from a directory and update the documents.

        **Args:**

        * **directory_path:** Path to the directory that contains the json files.
        * **batch_size:** The number of documents that should be included in each batch.
            Defaults to 1000.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.
        * **document_type:** The type of document being added. Accepted types are json, csv, and
                ndjson. For csv files the first row of the document should be a header row contining
                the field names, and ever for should have a title.
        * **combine_documents:** If set to True this will combine the documents from all the files
            before indexing them. Defaults to True.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **InvalidDocumentError:** If the docucment is not a valid format for MeiliSarch.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> directory_path = Path("/path/to/directory/containing/files")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_from_directory_in_batches(directory_path)
        ```
        """
        directory = Path(directory_path) if isinstance(directory_path, str) else directory_path

        if combine_documents:
            all_documents = []
            for path in directory.iterdir():
                if path.suffix == f".{document_type}":
                    documents = await Index._load_documents_from_file(path)
                    all_documents.append(documents)

            Index._raise_on_no_documents(all_documents, document_type, directory_path)

            loop = get_running_loop()
            combined = await loop.run_in_executor(
                None, partial(Index._combine_documents, all_documents)
            )

            return await self.update_documents_in_batches(
                combined, batch_size=batch_size, primary_key=primary_key
            )

        responses: list[TaskId] = []

        update_documents = []
        for path in directory.iterdir():
            if path.suffix == f".{document_type}":
                documents = await Index._load_documents_from_file(path)
                update_documents.append(
                    self.update_documents_in_batches(
                        documents, batch_size=batch_size, primary_key=primary_key
                    )
                )

        Index._raise_on_no_documents(update_documents, document_type, directory_path)

        if len(update_documents) > 1:
            # Send the first document on its own before starting the gather. Otherwise MeiliSearch
            # returns an error because it thinks all entries are trying to create the same index.
            first_response = await update_documents.pop()
            responses_gather = await gather(*update_documents)
            responses = [*first_response, *[x for y in responses_gather for x in y]]
        else:
            responses = await update_documents[0]

        return responses

    async def update_documents_from_file(
        self, file_path: Path | str, primary_key: str | None = None
    ) -> TaskId:
        """Add documents in the index from a json file.

        **Args:**

        * **file_path:** Path to the json file.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> file_path = Path("/path/to/file.json")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_from_file(file_path)
        ```
        """
        documents = await Index._load_documents_from_file(file_path)

        return await self.update_documents(documents, primary_key=primary_key)

    async def update_documents_from_file_auto_batch(
        self,
        file_path: Path | str,
        *,
        max_payload_size: int = 104857600,
        primary_key: str | None = None,
    ) -> list[TaskId]:
        """Automatically splits the documents into batches when updating from a json file.

        Each batch tries to fill the max_payload_size

        **Args:**

        * **file_path:** Path to the json file.
        * **batch_size:** The number of documents that should be included in each batch.
            Defaults to 1000.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> file_path = Path("/path/to/file.json")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_from_file_auto_batch(file_path)
        ```
        """
        documents = await Index._load_documents_from_file(file_path)

        task_ids = []
        async for batch in Index._generate_auto_batches(documents, max_payload_size):
            task_id = await self.update_documents(batch, primary_key)
            task_ids.append(task_id)

        return task_ids

    async def update_documents_from_file_in_batches(
        self, file_path: Path | str, *, batch_size: int = 1000, primary_key: str | None = None
    ) -> list[TaskId]:
        """Updates documents form a json file in batches to reduce RAM usage with indexing.

        **Args:**

        * **file_path:** Path to the json file.
        * **batch_size:** The number of documents that should be included in each batch.
            Defaults to 1000.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> file_path = Path("/path/to/file.json")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_from_file_in_batches(file_path)
        ```
        """
        documents = await Index._load_documents_from_file(file_path)

        return await self.update_documents_in_batches(
            documents, batch_size=batch_size, primary_key=primary_key
        )

    async def update_documents_from_raw_file(
        self, file_path: Path | str, primary_key: str | None = None
    ) -> TaskId:
        """Directly send csv or ndjson files to MeiliSearch without pre-processing.

        The can reduce RAM usage from MeiliSearch during indexing, but does not include the option
        for batching.

        **Args:**

        * **file_path:** The path to the file to send to MeiliSearch. Only csv and ndjson files are
            allowed.
        * **primary_key:** The primary key of the documents. This will be ignored if already set.
            Defaults to None.

        **Returns:** Update id to track the action.

        **Raises:**

        * **ValueError:** If the file is not a csv or ndjson file.
        * **MeiliSearchError:** If the file path is not valid
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from pathlib import Path
        >>> from meilisearch_async_client import Client
        >>> file_path = Path("/path/to/file.csv")
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_documents_from_raw_file(file_path)
        ```
        """
        upload_path = Path(file_path) if isinstance(file_path, str) else file_path
        if not upload_path.exists():
            raise MeiliSearchError("No file found at the specified path")

        if upload_path.suffix not in (".csv", ".ndjson"):
            raise ValueError("Only csv and ndjson files can be sent as binary files")

        content_type = "text/csv" if upload_path.suffix == ".csv" else "application/x-ndjson"
        url = self._documents_url
        if primary_key:
            formatted_primary_key = urlencode({"primaryKey": primary_key})
            url = f"{url}?{formatted_primary_key}"

        async with aiofiles.open(upload_path, "r") as f:
            data = await f.read()

        response = await self._http_requests.put(url, body=data, content_type=content_type)

        return TaskId(**response.json())

    async def delete_document(self, document_id: str) -> TaskId:
        """Delete one document from the index.

        **Args:**

        * **document_id:** Unique identifier of the document.

        **Returns:** Update id to track the action.

        **Rases:**

        * **MeiliSearchCommunicationError:** If there was an error communicating with the server.
        * **MeiliSearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.delete_document("1234")
        ```
        """
        url = f"{self._documents_url}/{document_id}"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def delete_documents(self, ids: list[str]) -> TaskId:
        """Delete multiple documents from the index.

        **Args:**

        * **ids:** List of unique identifiers of documents.

        **Returns:** List of update ids to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.delete_documents(["1234", "5678"])
        ```
        """
        url = f"{self._documents_url}/delete-batch"
        response = await self._http_requests.post(url, ids)

        return TaskId(**response.json())

    async def delete_all_documents(self) -> TaskId:
        """Delete all documents from the index.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.delete_all_document()
        ```
        """
        url = f"{self._documents_url}"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_settings(self) -> MeiliSearchSettings:
        """Get settings of the index.

        **Returns:** Settings of the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     settings = await index.get_settings()
        ```
        """
        url = f"{self._settings_url}"
        response = await self._http_requests.get(url)

        return MeiliSearchSettings(**response.json())

    async def update_settings(self, body: MeiliSearchSettings) -> TaskId:
        """Update settings of the index.

        **Args:**

        * **body:** Settings of the index.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> from meilisearch_async_client import MeiliSearchSettings
        >>> new_settings = MeiliSearchSettings(
        >>>     synonyms={"wolverine": ["xmen", "logan"], "logan": ["wolverine"]},
        >>>     stop_words=["the", "a", "an"],
        >>>     ranking_rules=[
        >>>         "words",
        >>>         "typo",
        >>>         "proximity",
        >>>         "attribute",
        >>>         "sort",
        >>>         "exactness",
        >>>         "release_date:desc",
        >>>         "rank:desc",
        >>>    ],
        >>>    filterable_attributes=["genre", "director"],
        >>>    distinct_attribute="url",
        >>>    searchable_attributes=["title", "description", "genre"],
        >>>    displayed_attributes=["title", "description", "genre", "release_date"],
        >>>    sortable_attributes=["title", "release_date"],
        >>> )
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_settings(new_settings)
        ```
        """
        body_dict = {k: v for k, v in body.dict(by_alias=True).items() if v is not None}

        url = f"{self._settings_url}"
        response = await self._http_requests.post(url, body_dict)

        return TaskId(**response.json())

    async def reset_settings(self) -> TaskId:
        """Reset settings of the index to default values.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_settings()
        ```
        """
        url = f"{self._settings_url}"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_ranking_rules(self) -> list[str]:
        """Get ranking rules of the index.

        **Returns:** List containing the ranking rules of the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     ranking_rules = await index.get_ranking_rules()
        ```
        """
        url = f"{self._settings_url}/ranking-rules"
        response = await self._http_requests.get(url)

        return response.json()

    async def update_ranking_rules(self, ranking_rules: list[str]) -> TaskId:
        """Update ranking rules of the index.

        **Args:**

        * **ranking_rules:** List containing the ranking rules.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> ranking_rules=[
        >>>      "words",
        >>>      "typo",
        >>>      "proximity",
        >>>      "attribute",
        >>>      "sort",
        >>>      "exactness",
        >>>      "release_date:desc",
        >>>      "rank:desc",
        >>> ],
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_ranking_rules(ranking_rules)
        ```
        """
        url = f"{self._settings_url}/ranking-rules"
        respose = await self._http_requests.post(url, ranking_rules)

        return TaskId(**respose.json())

    async def reset_ranking_rules(self) -> TaskId:
        """Reset ranking rules of the index to default values.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_ranking_rules()
        ```
        """
        url = f"{self._settings_url}/ranking-rules"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_distinct_attribute(self) -> str | None:
        """Get distinct attribute of the index.

        **Returns:** String containing the distinct attribute of the index. If no distinct attribute
            `None` is returned.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     distinct_attribute = await index.get_distinct_attribute()
        ```
        """
        url = f"{self._settings_url}/distinct-attribute"
        response = await self._http_requests.get(url)

        if not response.json():
            None

        return response.json()

    async def update_distinct_attribute(self, body: str) -> TaskId:
        """Update distinct attribute of the index.

        **Args:**

        * **body:** Distinct attribute.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_distinct_attribute("url")
        ```
        """
        url = f"{self._settings_url}/distinct-attribute"
        response = await self._http_requests.post(url, body)

        return TaskId(**response.json())

    async def reset_distinct_attribute(self) -> TaskId:
        """Reset distinct attribute of the index to default values.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_distinct_attributes()
        ```
        """
        url = f"{self._settings_url}/distinct-attribute"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_searchable_attributes(self) -> list[str]:
        """Get searchable attributes of the index.

        **Returns:** List containing the searchable attributes of the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     searchable_attributes = await index.get_searchable_attributes()
        ```
        """
        url = f"{self._settings_url}/searchable-attributes"
        response = await self._http_requests.get(url)
        return response.json()

    async def update_searchable_attributes(self, body: list[str]) -> TaskId:
        """Update searchable attributes of the index.

        **Args:**

        * **body:** List containing the searchable attributes.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_searchable_attributes(["title", "description", "genre"])
        ```
        """
        url = f"{self._settings_url}/searchable-attributes"
        response = await self._http_requests.post(url, body)

        return TaskId(**response.json())

    async def reset_searchable_attributes(self) -> TaskId:
        """Reset searchable attributes of the index to default values.

        **Args:**

        * **body:** List containing the searchable attributes.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_searchable_attributes()
        ```
        """
        url = f"{self._settings_url}/searchable-attributes"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_displayed_attributes(self) -> list[str]:
        """Get displayed attributes of the index.

        **Returns:** List containing the displayed attributes of the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     displayed_attributes = await index.get_displayed_attributes()
        ```
        """
        url = f"{self._settings_url}/displayed-attributes"
        response = await self._http_requests.get(url)
        return response.json()

    async def update_displayed_attributes(self, body: list[str]) -> TaskId:
        """Update displayed attributes of the index.

        **Args:**

        * **body:** List containing the displayed attributes.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_displayed_attributes(
        >>>         ["title", "description", "genre", "release_date"]
        >>>     )
        ```
        """
        url = f"{self._settings_url}/displayed-attributes"
        response = await self._http_requests.post(url, body)

        return TaskId(**response.json())

    async def reset_displayed_attributes(self) -> TaskId:
        """Reset displayed attributes of the index to default values.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_displayed_attributes()
        ```
        """
        url = f"{self._settings_url}/displayed-attributes"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_stop_words(self) -> list[str] | None:
        """Get stop words of the index.

        **Returns:** List containing the stop words of the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     stop_words = await index.get_stop_words()
        ```
        """
        url = f"{self._settings_url}/stop-words"
        response = await self._http_requests.get(url)

        if not response.json():
            return None

        return response.json()

    async def update_stop_words(self, body: list[str]) -> TaskId:
        """Update stop words of the index.

        **Args:**

        * **body:** List containing the stop words of the index.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_stop_words(["the", "a", "an"])
        ```
        """
        url = f"{self._settings_url}/stop-words"
        response = await self._http_requests.post(url, body)

        return TaskId(**response.json())

    async def reset_stop_words(self) -> TaskId:
        """Reset stop words of the index to default values.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_stop_words()
        ```
        """
        url = f"{self._settings_url}/stop-words"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_synonyms(self) -> dict[str, list[str]] | None:
        """Get synonyms of the index.

        **Returns:** The synonyms of the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     synonyms = await index.get_synonyms()
        ```
        """
        url = f"{self._settings_url}/synonyms"
        response = await self._http_requests.get(url)

        if not response.json():
            return None

        return response.json()

    async def update_synonyms(self, body: dict[str, list[str]]) -> TaskId:
        """Update synonyms of the index.

        **Args:**

        * **body:** The synonyms of the index.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_synonyms(
        >>>         {"wolverine": ["xmen", "logan"], "logan": ["wolverine"]}
        >>>     )
        ```
        """
        url = f"{self._settings_url}/synonyms"
        response = await self._http_requests.post(url, body)

        return TaskId(**response.json())

    async def reset_synonyms(self) -> TaskId:
        """Reset synonyms of the index to default values.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_synonyms()
        ```
        """
        url = f"{self._settings_url}/synonyms"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_filterable_attributes(self) -> list[str] | None:
        """Get filterable attributes of the index.

        **Returns:** List containing the filterable attributes of the index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     filterable_attributes = await index.get_filterable_attributes()
        ```
        """
        url = f"{self._settings_url}/filterable-attributes"
        response = await self._http_requests.get(url)

        if not response.json():
            return None

        return response.json()

    async def update_filterable_attributes(self, body: list[str]) -> TaskId:
        """Update filterable attributes of the index.

        **Args:**

        * **body:** List containing the filterable attributes of the index.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_filterable_attributes(["genre", "director"])
        ```
        """
        url = f"{self._settings_url}/filterable-attributes"
        response = await self._http_requests.post(url, body)

        return TaskId(**response.json())

    async def reset_filterable_attributes(self) -> TaskId:
        """Reset filterable attributes of the index to default values.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_filterable_attributes()
        ```
        """
        url = f"{self._settings_url}/filterable-attributes"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    async def get_sortable_attributes(self) -> list[str]:
        """Get sortable attributes of the Index.

        **Args:**

        * **sortable_attributes:** List containing the sortable attributes of the index.

        **Returns:** List containing the sortable attributes of the Index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     sortable_attributes = await index.get_sortable_attributes()
        ```
        """
        url = f"{self._settings_url}/sortable-attributes"
        response = await self._http_requests.get(url)

        return response.json()

    async def update_sortable_attributes(self, sortable_attributes: list[str]) -> TaskId:
        """Get sortable attributes of the Index.

        **Returns:** List containing the sortable attributes of the Index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.update_sortable_attributes(["title", "release_date"])
        ```
        """
        url = f"{self._settings_url}/sortable-attributes"
        response = await self._http_requests.post(url, sortable_attributes)

        return TaskId(**response.json())

    async def reset_sortable_attributes(self) -> TaskId:
        """Reset sortable attributes of the index to default values.

        **Returns:** Update id to track the action.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        >>>     await index.reset_sortable_attributes()
        ```
        """
        url = f"{self._settings_url}/sortable-attributes"
        response = await self._http_requests.delete(url)

        return TaskId(**response.json())

    @staticmethod
    async def _batch(documents: list[dict], batch_size: int) -> AsyncGenerator[list[dict], None]:
        total_len = len(documents)
        for i in range(0, total_len, batch_size):
            yield documents[i : i + batch_size]

    @staticmethod
    def _raise_on_no_documents(
        documents: list[Any], document_type: str, directory_path: str | Path
    ) -> None:
        if not documents:
            raise MeiliSearchError(f"No {document_type} files found in {directory_path}")

    @staticmethod
    def _combine_documents(documents: list[list[Any]]) -> list[Any]:
        return [x for y in documents for x in y]

    @staticmethod
    async def _generate_auto_batches(
        documents: list[dict[str, Any]], max_payload_size: int
    ) -> AsyncGenerator[list[dict], None]:
        loop = get_running_loop()

        # Check the size of all documents together if it is below the max size yield it all at onece
        doc_json_str = await loop.run_in_executor(None, partial(json.dumps, documents))
        doc_size = await loop.run_in_executor(None, partial(getsizeof, doc_json_str))
        if doc_size < max_payload_size:
            yield documents
        else:
            batch = []
            batch_size = 0
            for doc in documents:
                doc_json_str = await loop.run_in_executor(None, partial(json.dumps, doc))
                doc_size = await loop.run_in_executor(None, partial(getsizeof, doc_json_str))
                if doc_size > max_payload_size:
                    raise PayloadTooLarge(
                        f"Payload size {doc_size} is greater than the maximum payload size of {max_payload_size}"
                    )
                batch_size += doc_size
                batch.append(doc)
                if batch_size >= max_payload_size:
                    batch.pop()
                    yield batch
                    batch.clear()
                    batch.append(doc)
                    batch_size = doc_size
            if batch:
                yield batch

    @staticmethod
    def _iso_to_date_time(iso_date: datetime | str | None) -> datetime | None:
        """Handle conversion of iso string to datetime.

        The microseconds from MeiliSearch are sometimes too long for python to convert so this
        strips off the last digits to shorten it when that happens.
        """
        if not iso_date:
            return None

        if isinstance(iso_date, datetime):
            return iso_date

        try:
            return datetime.strptime(iso_date, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            split = iso_date.split(".")
            reduce = len(split[1]) - 6
            reduced = f"{split[0]}.{split[1][:-reduce]}Z"
            return datetime.strptime(reduced, "%Y-%m-%dT%H:%M:%S.%fZ")

    @staticmethod
    async def _load_documents_from_file(
        file_path: Path | str,
    ) -> list[dict[Any, Any]]:
        if isinstance(file_path, str):
            file_path = Path(file_path)

        loop = get_running_loop()
        await loop.run_in_executor(None, partial(Index._validate_file_type, file_path))

        if file_path.suffix == ".csv":
            with open(file_path, mode="r") as f:
                documents = await loop.run_in_executor(None, partial(DictReader, f))
                return list(documents)

        if file_path.suffix == ".ndjson":
            with open(file_path, mode="r") as f:
                return [await loop.run_in_executor(None, partial(json.loads, x)) for x in f]

        async with aiofiles.open(file_path, mode="r") as f:  # type: ignore
            data = await f.read()  # type: ignore
            documents = await loop.run_in_executor(None, partial(json.loads, data))

            if not isinstance(documents, list):
                raise InvalidDocumentError("MeiliSearch requires documents to be in a list")

            return documents

    @staticmethod
    def _validate_file_type(file_path: Path) -> None:
        if file_path.suffix not in (".json", ".csv", ".ndjson"):
            raise MeiliSearchError("File must be a json, ndjson, or csv file")



from datetime import datetime
from typing import Dict, List, Optional

from camel_converter.pydantic_base import CamelBase

from meilisearch_python_async.models.index import IndexStats


class ClientStats(CamelBase):
    database_size: int
    last_update: Optional[datetime] = None
    indexes: Optional[Dict[str, IndexStats]] = None


class Key(CamelBase):
    description: str
    key: str
    actions: List[str]
    indexes: List[str]
    expires_at: Optional[datetime] = None
    created_at: datetime
    updated_at: Optional[datetime] = None


class KeyCreate(CamelBase):
    description: str
    actions: List[str]
    indexes: List[str]
    expires_at: Optional[datetime] = None

    class Config:
        json_encoders = {
            datetime: lambda v: None if not v else f"{str(v).split('.')[0].replace(' ', 'T')}Z"
        }


class KeyUpdate(CamelBase):
    key: str
    description: Optional[str] = None
    actions: Optional[List[str]] = None
    indexes: Optional[List[str]] = None
    expires_at: Optional[datetime] = None

    class Config:
        json_encoders = {
            datetime: lambda v: None if not v else f"{str(v).split('.')[0].replace(' ', 'T')}Z"
        }



from datetime import datetime
from typing import Any, Dict, Optional, Union

from camel_converter.pydantic_base import CamelBase
from pydantic import Field


class TaskId(CamelBase):
    uid: int


class TaskStatus(TaskId):
    index_uid: str
    status: str
    task_type: Union[str, Dict[str, Any]] = Field(..., alias="type")
    details: Optional[Dict[str, Any]]
    duration: Optional[str]
    enqueued_at: datetime
    started_at: Optional[datetime]
    finished_at: Optional[datetime]



from datetime import datetime
from typing import Union

from camel_converter.pydantic_base import CamelBase


class Version(CamelBase):
    commit_sha: str
    commit_date: Union[datetime, str]
    pkg_version: str



from __future__ import annotations

import json
from types import TracebackType
from typing import Type

from httpx import AsyncClient

from meilisearch_python_async._http_requests import HttpRequests
from meilisearch_python_async.errors import MeiliSearchApiError
from meilisearch_python_async.index import Index
from meilisearch_python_async.models.client import ClientStats, Key, KeyCreate, KeyUpdate
from meilisearch_python_async.models.dump import DumpInfo
from meilisearch_python_async.models.health import Health
from meilisearch_python_async.models.index import IndexInfo
from meilisearch_python_async.models.version import Version
from meilisearch_python_async.task import wait_for_task


class Client:
    """The client to connect to the MeiliSearchApi."""

    def __init__(self, url: str, api_key: str | None = None, *, timeout: int | None = None) -> None:
        """Class initializer.

        **Args:**

        * **url:** The url to the MeiliSearch API (ex: http://localhost:7700)
        * **api_key:** The optional API key for MeiliSearch. Defaults to None.
        * **timeout:** The amount of time in seconds that the client will wait for a response before
            timing out. Defaults to None.
        """
        if api_key:
            headers = {"Authorization": f"Bearer {api_key}"}
        else:
            headers = None

        self.http_client = AsyncClient(base_url=url, timeout=timeout, headers=headers)
        self._http_requests = HttpRequests(self.http_client)

    async def __aenter__(self) -> Client:
        return self

    async def __aexit__(
        self,
        et: Type[BaseException] | None,
        ev: Type[BaseException] | None,
        traceback: TracebackType | None,
    ) -> None:
        await self.aclose()

    async def aclose(self) -> None:
        """Closes the client.

        This only needs to be used if the client was not created with a context manager.
        """
        await self.http_client.aclose()

    async def create_dump(self) -> DumpInfo:
        """Trigger the creation of a MeiliSearch dump.

        **Returns:** Information about the dump.
            https://docs.meilisearch.com/reference/api/dump.html#create-a-dump

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     await client.create_dump()
        ```
        """
        response = await self._http_requests.post("dumps")
        return DumpInfo(**response.json())

    async def create_index(self, uid: str, primary_key: str | None = None) -> Index:
        """Creates a new index.

        **Args:**

        * **uid:** The index's unique identifier.
        * **primary_key:** The primary key of the documents. Defaults to None.

        **Returns:** An instance of Index containing the information of the newly created index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = await client.create_index("movies")
        ```
        """
        return await Index.create(self.http_client, uid, primary_key)

    async def delete_index_if_exists(self, uid: str) -> bool:
        """Deletes an index if it already exists.

        **Args:**

        * **uid:** The index's unique identifier.

        **Returns:** True if an index was deleted for False if not.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     await client.delete_index_if_exists()
        ```
        """
        url = f"indexes/{uid}"
        response = await self._http_requests.delete(url)
        status = await wait_for_task(self.http_client, response.json()["uid"])
        if status.status == "succeeded":
            return True
        return False

    async def get_indexes(self) -> list[Index] | None:
        """Get all indexes.

        **Returns:** A list of all indexes.

        **Raises:**
        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     indexes = await client.get_indexes()
        ```
        """
        response = await self._http_requests.get("indexes")

        if not response.json():
            return None

        return [
            Index(
                http_client=self.http_client,
                uid=x["uid"],
                primary_key=x["primaryKey"],
                created_at=x["createdAt"],
                updated_at=x["updatedAt"],
            )
            for x in response.json()
        ]

    async def get_index(self, uid: str) -> Index:
        """Gets a single index based on the uid of the index.

        **Args:**

        * **uid:** The index's unique identifier.

        **Returns:** An Index instance containing the information of the fetched index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = await client.get_index()
        ```
        """
        return await Index(self.http_client, uid).fetch_info()

    def index(self, uid: str) -> Index:
        """Create a local reference to an index identified by UID, without making an HTTP call.

        Because no network call is made this method is not awaitable.

        **Args:**

        * **uid:** The index's unique identifier.

        **Returns:** An Index instance.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = client.index("movies")
        ```
        """
        return Index(self.http_client, uid=uid)

    async def get_all_stats(self) -> ClientStats:
        """Get stats for all indexes.

        **Returns:** Information about database size and all indexes.
            https://docs.meilisearch.com/reference/api/stats.html

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     stats = await client.get_all_stats()
        ```
        """
        response = await self._http_requests.get("stats")
        return ClientStats(**response.json())

    async def get_dump_status(self, uid: str) -> DumpInfo:
        """Retrieve the status of a MeiliSearch dump creation.

        **Args:**
        * **uid:** The update identifier for the dump creation.

        **Returns:** Information about the dump status.
            https://docs.meilisearch.com/reference/api/dump.html#get-dump-status

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     status = await client.get_dump_status("20201101-110357260")
        ```
        """
        url = f"dumps/{uid}/status"
        response = await self._http_requests.get(url)
        return DumpInfo(**response.json())

    async def get_or_create_index(self, uid: str, primary_key: str | None = None) -> Index:
        """Get an index, or create it if it doesn't exist.

        **Args:**

        * **uid:** The index's unique identifier.
        * **primary_key:** The primary key of the documents. Defaults to None.

        **Returns:** An instance of Index containing the information of the retrieved or newly created index.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.MeiliSearchTimeoutError: If the connection times out.
        * **MeiliSearchTimeoutError:** If the connection times out.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = await client.get_or_create_index("movies")
        ```
        """
        try:
            index_instance = await self.get_index(uid)
        except MeiliSearchApiError as err:
            if "index_not_found" not in err.code:
                raise
            index_instance = await self.create_index(uid, primary_key)
        return index_instance

    async def create_key(self, key: KeyCreate) -> Key:
        """Creates a new API key.

        **Args:**

        * **key:** The information to use in creating the key. Note that if an expires_at value
            is included it should be in UTC time.

        **Returns:** The new API key.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> from meilissearch_async_client.models.client import KeyCreate
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     key_info = KeyCreate(
        >>>         description="My new key",
        >>>         actions=["search"],
        >>>         indexes=["movies"],
        >>>     )
        >>>     keys = await client.create_key(key_info)
        ```
        """
        # The json.loads(key.json()) is because Pydantic can't serialize a date in a Python dict,
        # but can when converting to a json string.
        response = await self._http_requests.post("keys", json.loads(key.json(by_alias=True)))
        return Key(**response.json())

    async def delete_key(self, key: str) -> int:
        """Deletes an API key.

        **Args:**

        * **key:** The key to delete.

        **Returns:** The Response status code. 204 signifies a successful delete.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     await client.delete_key("abc123")
        ```
        """
        response = await self._http_requests.delete(f"keys/{key}")
        return response.status_code

    async def get_keys(self) -> list[Key]:
        """Gets the MeiliSearch API keys.

        **Returns:** API keys.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     keys = await client.get_keys()
        ```
        """
        response = await self._http_requests.get("keys")
        keys = [Key(**x) for x in response.json()["results"]]

        return keys

    async def get_key(self, key: str) -> Key:
        """Gets information about a specific API key.

        **Args:**

        * **key:** The key for which to retrieve the information.

        **Returns:** The API key, or `None` if the key is not found.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     keys = await client.get_key("abc123")
        ```
        """
        response = await self._http_requests.get(f"keys/{key}")
        return Key(**response.json())

    async def update_key(self, key: KeyUpdate) -> Key:
        """Update an API key.

        **Args:**

        * **key:** The information to use in updating the key. Note that if an expires_at value
            is included it should be in UTC time.

        **Returns:** The updated API key.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> from meilissearch_async_client.models.client import KeyUpdate
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     key_info = KeyUpdate(
                    key="abc123",
        >>>         indexes=["*"],
        >>>     )
        >>>     keys = await client.update_key(key_info)
        ```
        """
        # The json.loads(key.json()) is because Pydantic can't serialize a date in a Python dict,
        # but can when converting to a json string.
        response = await self._http_requests.patch(
            f"keys/{key.key}", json.loads(key.json(by_alias=True))
        )
        return Key(**response.json())

    async def get_raw_index(self, uid: str) -> IndexInfo | None:
        """Gets the index and returns all the index information rather than an Index instance.

        **Args:**

        * **uid:** The index's unique identifier.

        **Returns:** Index information rather than an Index instance.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = await client.get_raw_index("movies")
        ```
        """
        response = await self.http_client.get(f"indexes/{uid}")

        if response.status_code == 404:
            return None

        return IndexInfo(**response.json())

    async def get_raw_indexes(self) -> list[IndexInfo] | None:
        """Gets all the indexes.

        Returns all the index information rather than an Index instance.

        **Returns:** A list of the Index information rather than an Index instances.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     index = await client.get_raw_indexes()
        ```
        """
        response = await self._http_requests.get("indexes")

        if not response.json():
            return None

        return [IndexInfo(**x) for x in response.json()]

    async def get_version(self) -> Version:
        """Get the MeiliSearch version.

        **Returns:** Information about the version of MeiliSearch.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     version = await client.get_version()
        ```
        """
        response = await self._http_requests.get("version")

        return Version(**response.json())

    async def health(self) -> Health:
        """Get health of the MeiliSearch server.

        **Returns:** The status of the MeiliSearch server.

        **Raises:**

        * **MeilisearchCommunicationError:** If there was an error communicating with the server.
        * **MeilisearchApiError:** If the MeiliSearch API returned an error.

        Usage:

        ```py
        >>> from meilisearch_async_client import Client
        >>> async with Client("http://localhost.com", "masterKey") as client:
        >>>     health = await client.get_healths()
        ```
        """
        response = await self._http_requests.get("health")
        return Health(**response.json())



from __future__ import annotations

from asyncio import sleep
from datetime import datetime

from httpx import AsyncClient

from meilisearch_python_async._http_requests import HttpRequests
from meilisearch_python_async.errors import MeiliSearchTimeoutError
from meilisearch_python_async.models.task import TaskStatus


async def get_tasks(http_client: AsyncClient, index_id: str | None = None) -> list[TaskStatus]:
    """Get all tasks.

    **Args:**

    * **http_client:** An AsyncClient instance.
    * **index_id:** The id of the index for which to get the tasks. If provided this will get the
        tasks only for the specified index, if not all tasks will be returned. Default = None

    **Returns:** A list of all tasks.

    **Raises:**

    * **MeilisearchCommunicationError:** If there was an error communicating with the server.
    * **MeilisearchApiError:** If the MeiliSearch API returned an error.
    * **MeiliSearchTimeoutError:** If the connection times out.

    Usage:

    ```py
    >>> from meilisearch_async_client import Client
    >>> from meilisearch_async_client.task import get_tasks
    >>>
    >>> async with Client("http://localhost.com", "masterKey") as client:
    >>>     await get_tasks(client.http_client)
    ```
    """
    url = f"indexes/{index_id}/tasks" if index_id else "tasks"
    response = await http_client.get(url)
    tasks = [TaskStatus(**x) for x in response.json()["results"]]

    return tasks


async def get_task(http_client: AsyncClient, task_id: int) -> TaskStatus:
    """Get a single task from it's task id.

    **Args:**

    * **http_client:** An AsyncClient instance.
    * **task_id:** Identifier of the task to retrieve.

    **Returns:** A list of all tasks.

    **Raises:**

    * **MeilisearchCommunicationError:** If there was an error communicating with the server.
    * **MeilisearchApiError:** If the MeiliSearch API returned an error.
    * **MeiliSearchTimeoutError:** If the connection times out.

    Usage:

    ```py
    >>> from meilisearch_async_client import Client
    >>> from meilisearch_async_client.task import get_task
    >>>
    >>> async with Client("http://localhost.com", "masterKey") as client:
    >>>     await get_task(client.http_client, 1244)
    ```
    """
    response = await http_client.get(f"tasks/{task_id}")

    return TaskStatus(**response.json())


async def wait_for_task(
    http_client: AsyncClient, task_id: int, *, timeout_in_ms: int = 5000, interval_in_ms: int = 50
) -> TaskStatus:
    """Wait until MeiliSearch processes a task, and get its status.

    **Args:**

    * **http_client:** An AsyncClient instance.
    * **task_id:** Identifier of the task to retrieve.
    * **timeout_in_ms:** Amount of time in milliseconds to wait before raising a
        MeiliSearchTimeoutError. Defaults to 5000.
    * **interval_in_ms:** Time interval in miliseconds to sleep between requests. Defaults to 50.

    **Returns:** Details of the processed update status.

    **Raises:**

    * **MeilisearchCommunicationError:** If there was an error communicating with the server.
    * **MeilisearchApiError:** If the MeiliSearch API returned an error.
    * **MeiliSearchTimeoutError:** If the connection times out.

    Usage:

    ```py
    >>> from meilisearch_async_client import Client
    >>> from meilisearch_async_client.task import wait_for_task
    >>> >>> documents = [
    >>>     {"id": 1, "title": "Movie 1", "genre": "comedy"},
    >>>     {"id": 2, "title": "Movie 2", "genre": "drama"},
    >>> ]
    >>> async with Client("http://localhost.com", "masterKey") as client:
    >>>     index = client.index("movies")
    >>>     response = await index.add_documents(documents)
    >>>     await wait_for_pending_task(client.http_client, response.update_id)
    ```
    """
    url = f"tasks/{task_id}"
    http_requests = HttpRequests(http_client)
    start_time = datetime.now()
    elapsed_time = 0.0
    while elapsed_time < timeout_in_ms:
        response = await http_requests.get(url)
        status = TaskStatus(**response.json())
        if status.status in ("succeeded", "failed"):
            return status
        await sleep(interval_in_ms / 1000)
        time_delta = datetime.now() - start_time
        elapsed_time = time_delta.seconds * 1000 + time_delta.microseconds / 1000
    raise MeiliSearchTimeoutError(
        f"timeout of {timeout_in_ms}ms has exceeded on process {task_id} when waiting for pending update to resolve."
    )
